package tests

import (
	"testing"

	assert "git.sharkk.net/Go/Assert"
	"git.sharkk.net/Sharkk/Mako/types"
	"git.sharkk.net/Sharkk/Mako/vm"
)

func TestVMPushPop(t *testing.T) {
	vm := vm.New()

	// Create basic bytecode that pushes constants and then pops
	bytecode := &types.Bytecode{
		Constants: []any{5.0, "hello", true},
		Instructions: []types.Instruction{
			{Opcode: types.OpConstant, Operand: 0}, // Push 5.0
			{Opcode: types.OpConstant, Operand: 1}, // Push "hello"
			{Opcode: types.OpConstant, Operand: 2}, // Push true
			{Opcode: types.OpPop, Operand: 0},      // Pop true
			{Opcode: types.OpPop, Operand: 0},      // Pop "hello"
			{Opcode: types.OpPop, Operand: 0},      // Pop 5.0
		},
	}

	// Run the VM
	vm.Run(bytecode)

	// Check stack is empty
	assert.Equal(t, 0, len(vm.CurrentStack()))
}

func TestVMArithmetic(t *testing.T) {
	tests := []struct {
		constants    []any
		instructions []types.Instruction
		expected     float64
	}{
		{
			// 5 + 10
			[]any{5.0, 10.0},
			[]types.Instruction{
				{Opcode: types.OpConstant, Operand: 0}, // Push 5.0
				{Opcode: types.OpConstant, Operand: 1}, // Push 10.0
				{Opcode: types.OpAdd, Operand: 0},      // Add
			},
			15.0,
		},
		{
			// 5 - 10
			[]any{5.0, 10.0},
			[]types.Instruction{
				{Opcode: types.OpConstant, Operand: 0}, // Push 5.0
				{Opcode: types.OpConstant, Operand: 1}, // Push 10.0
				{Opcode: types.OpSubtract, Operand: 0}, // Subtract
			},
			-5.0,
		},
		{
			// 5 * 10
			[]any{5.0, 10.0},
			[]types.Instruction{
				{Opcode: types.OpConstant, Operand: 0}, // Push 5.0
				{Opcode: types.OpConstant, Operand: 1}, // Push 10.0
				{Opcode: types.OpMultiply, Operand: 0}, // Multiply
			},
			50.0,
		},
		{
			// 10 / 5
			[]any{10.0, 5.0},
			[]types.Instruction{
				{Opcode: types.OpConstant, Operand: 0}, // Push 10.0
				{Opcode: types.OpConstant, Operand: 1}, // Push 5.0
				{Opcode: types.OpDivide, Operand: 0},   // Divide
			},
			2.0,
		},
		{
			// -5
			[]any{5.0},
			[]types.Instruction{
				{Opcode: types.OpConstant, Operand: 0}, // Push 5.0
				{Opcode: types.OpNegate, Operand: 0},   // Negate
			},
			-5.0,
		},
	}

	for _, tt := range tests {
		vm := vm.New()
		bytecode := &types.Bytecode{
			Constants:    tt.constants,
			Instructions: tt.instructions,
		}

		vm.Run(bytecode)

		// Check the result on the stack
		stack := vm.CurrentStack()
		assert.Equal(t, 1, len(stack))
		assert.Equal(t, types.TypeNumber, stack[0].Type)
		assert.Equal(t, tt.expected, stack[0].Data.(float64))
	}
}

func TestVMComparisons(t *testing.T) {
	tests := []struct {
		constants    []any
		instructions []types.Instruction
		expected     bool
	}{
		{
			// 5 == 5
			[]any{5.0, 5.0},
			[]types.Instruction{
				{Opcode: types.OpConstant, Operand: 0}, // Push 5.0
				{Opcode: types.OpConstant, Operand: 1}, // Push 5.0
				{Opcode: types.OpEqual, Operand: 0},    // Equal
			},
			true,
		},
		{
			// 5 != 10
			[]any{5.0, 10.0},
			[]types.Instruction{
				{Opcode: types.OpConstant, Operand: 0}, // Push 5.0
				{Opcode: types.OpConstant, Operand: 1}, // Push 10.0
				{Opcode: types.OpNotEqual, Operand: 0}, // Not Equal
			},
			true,
		},
		{
			// 5 < 10
			[]any{5.0, 10.0},
			[]types.Instruction{
				{Opcode: types.OpConstant, Operand: 0}, // Push 5.0
				{Opcode: types.OpConstant, Operand: 1}, // Push 10.0
				{Opcode: types.OpLessThan, Operand: 0}, // Less Than
			},
			true,
		},
		{
			// 10 > 5
			[]any{10.0, 5.0},
			[]types.Instruction{
				{Opcode: types.OpConstant, Operand: 0},    // Push 10.0
				{Opcode: types.OpConstant, Operand: 1},    // Push 5.0
				{Opcode: types.OpGreaterThan, Operand: 0}, // Greater Than
			},
			true,
		},
		{
			// 5 <= 5
			[]any{5.0, 5.0},
			[]types.Instruction{
				{Opcode: types.OpConstant, Operand: 0},  // Push 5.0
				{Opcode: types.OpConstant, Operand: 1},  // Push 5.0
				{Opcode: types.OpLessEqual, Operand: 0}, // Less Than or Equal
			},
			true,
		},
		{
			// 5 >= 5
			[]any{5.0, 5.0},
			[]types.Instruction{
				{Opcode: types.OpConstant, Operand: 0},     // Push 5.0
				{Opcode: types.OpConstant, Operand: 1},     // Push 5.0
				{Opcode: types.OpGreaterEqual, Operand: 0}, // Greater Than or Equal
			},
			true,
		},
	}

	for _, tt := range tests {
		vm := vm.New()
		bytecode := &types.Bytecode{
			Constants:    tt.constants,
			Instructions: tt.instructions,
		}

		vm.Run(bytecode)

		// Check the result on the stack
		stack := vm.CurrentStack()
		assert.Equal(t, 1, len(stack))
		assert.Equal(t, types.TypeBoolean, stack[0].Type)
		assert.Equal(t, tt.expected, stack[0].Data.(bool))
	}
}

func TestVMTableOperations(t *testing.T) {
	// Create bytecode for: table = {}; table["key"] = "value"; result = table["key"]
	bytecode := &types.Bytecode{
		Constants: []any{"table", "key", "value", "result"},
		Instructions: []types.Instruction{
			{Opcode: types.OpNewTable, Operand: 0},  // Create new table
			{Opcode: types.OpSetGlobal, Operand: 0}, // Store in global "table"
			{Opcode: types.OpGetGlobal, Operand: 0}, // Push table onto stack
			{Opcode: types.OpConstant, Operand: 1},  // Push "key"
			{Opcode: types.OpConstant, Operand: 2},  // Push "value"
			{Opcode: types.OpSetIndex, Operand: 0},  // Set table["key"] = "value"
			{Opcode: types.OpPop, Operand: 0},       // Pop the result of SetIndex
			{Opcode: types.OpGetGlobal, Operand: 0}, // Push table onto stack again
			{Opcode: types.OpConstant, Operand: 1},  // Push "key"
			{Opcode: types.OpGetIndex, Operand: 0},  // Get table["key"]
			{Opcode: types.OpSetGlobal, Operand: 3}, // Store in global "result"
		},
	}

	vm := vm.New()
	vm.Run(bytecode)

	// Now check if "result" contains "value"
	result, found := vm.GetGlobal("result")
	assert.True(t, found)
	assert.Equal(t, types.TypeString, result.Type)
	assert.Equal(t, "value", result.Data.(string))
}

func TestVMConditionalJumps(t *testing.T) {
	// Create bytecode for: if true then result = "true" else result = "false" end
	bytecode := &types.Bytecode{
		Constants: []any{true, "result", "true", "false"},
		Instructions: []types.Instruction{
			{Opcode: types.OpConstant, Operand: 0},    // Push true
			{Opcode: types.OpJumpIfFalse, Operand: 6}, // Jump to else branch if false
			{Opcode: types.OpConstant, Operand: 2},    // Push "true"
			{Opcode: types.OpSetGlobal, Operand: 1},   // Set result = "true"
			{Opcode: types.OpJump, Operand: 8},        // Jump to end
			{Opcode: types.OpConstant, Operand: 3},    // Push "false"
			{Opcode: types.OpSetGlobal, Operand: 1},   // Set result = "false"
		},
	}

	vm := vm.New()
	vm.Run(bytecode)

	// Check result == "true"
	result, found := vm.GetGlobal("result")
	assert.True(t, found)
	assert.Equal(t, types.TypeString, result.Type)
	assert.Equal(t, "true", result.Data.(string))
}

func TestVMScopes(t *testing.T) {
	// Create bytecode for:
	// x = 5
	// {
	//   x = 10
	//   y = 20
	// }
	// result = x
	bytecode := &types.Bytecode{
		Constants: []any{5.0, "x", 10.0, 20.0, "y", "result"},
		Instructions: []types.Instruction{
			{Opcode: types.OpConstant, Operand: 0},   // Push 5.0
			{Opcode: types.OpSetGlobal, Operand: 1},  // Set global x = 5
			{Opcode: types.OpEnterScope, Operand: 0}, // Enter new scope
			{Opcode: types.OpConstant, Operand: 2},   // Push 10.0
			{Opcode: types.OpSetLocal, Operand: 1},   // Set local x = 10
			{Opcode: types.OpConstant, Operand: 3},   // Push 20.0
			{Opcode: types.OpSetLocal, Operand: 4},   // Set local y = 20
			{Opcode: types.OpExitScope, Operand: 0},  // Exit scope
			{Opcode: types.OpGetGlobal, Operand: 1},  // Get global x
			{Opcode: types.OpSetGlobal, Operand: 5},  // Set result = x
		},
	}

	vm := vm.New()
	vm.Run(bytecode)

	// Check result == 5.0 (global x, not the shadowed local x)
	result, found := vm.GetGlobal("result")
	assert.True(t, found)
	assert.Equal(t, types.TypeNumber, result.Type)
	assert.Equal(t, 5.0, result.Data.(float64))
}

func TestVMLogicalOperators(t *testing.T) {
	// Test NOT operator
	notBytecode := &types.Bytecode{
		Constants: []any{true, "result"},
		Instructions: []types.Instruction{
			{Opcode: types.OpConstant, Operand: 0},  // Push true
			{Opcode: types.OpNot, Operand: 0},       // NOT operation
			{Opcode: types.OpSetGlobal, Operand: 1}, // Set result
		},
	}

	vm := vm.New()
	vm.Run(notBytecode)

	// Check result == false
	result, found := vm.GetGlobal("result")
	assert.True(t, found)
	assert.Equal(t, types.TypeBoolean, result.Type)
	assert.Equal(t, false, result.Data.(bool))

	// Test AND with short-circuit
	andBytecode := &types.Bytecode{
		Constants: []any{false, 5.0, "x", true, "result"},
		Instructions: []types.Instruction{
			{Opcode: types.OpConstant, Operand: 1},    // Push 5.0
			{Opcode: types.OpSetGlobal, Operand: 2},   // Set x = 5.0
			{Opcode: types.OpConstant, Operand: 0},    // Push false
			{Opcode: types.OpDup, Operand: 0},         // Duplicate false for condition
			{Opcode: types.OpJumpIfFalse, Operand: 9}, // Jump if false (short-circuit)
			{Opcode: types.OpPop, Operand: 0},         // Pop the duplicate
			{Opcode: types.OpGetGlobal, Operand: 2},   // Get x
			{Opcode: types.OpConstant, Operand: 1},    // Push 5.0
			{Opcode: types.OpAdd, Operand: 0},         // Add x + 5 (should be skipped)
			{Opcode: types.OpSetGlobal, Operand: 4},   // Set result (this is where we jump to)
		},
	}

	vm = vm.Reset()
	vm.Run(andBytecode)

	// Check result == false (the first operand, since AND short-circuits)
	result, found = vm.GetGlobal("result")
	assert.True(t, found)
	assert.Equal(t, types.TypeBoolean, result.Type)
	assert.Equal(t, false, result.Data.(bool))

	// Test OR with short-circuit
	orBytecode := &types.Bytecode{
		Constants: []any{true, 5.0, "x", false, "result"},
		Instructions: []types.Instruction{
			{Opcode: types.OpConstant, Operand: 1},    // Push 5.0
			{Opcode: types.OpSetGlobal, Operand: 2},   // Set x = 5.0
			{Opcode: types.OpConstant, Operand: 0},    // Push true
			{Opcode: types.OpDup, Operand: 0},         // Duplicate true for condition
			{Opcode: types.OpJumpIfFalse, Operand: 7}, // Jump if false (not taken)
			{Opcode: types.OpConstant, Operand: 0},    // Push true again
			{Opcode: types.OpSetGlobal, Operand: 4},   // Set result (with actual value)
			{Opcode: types.OpJump, Operand: 11},       // Jump to end (short-circuit)
			{Opcode: types.OpPop, Operand: 0},         // Pop the duplicate (not executed)
			{Opcode: types.OpGetGlobal, Operand: 2},   // Get x (not executed)
			{Opcode: types.OpConstant, Operand: 1},    // Push 5.0 (not executed)
			{Opcode: types.OpAdd, Operand: 0},         // Add x + 5 (not executed)
		},
	}

	vm = vm.Reset()
	vm.Run(orBytecode)

	// Check result == true (the first operand, since OR short-circuits)
	result, found = vm.GetGlobal("result")
	assert.True(t, found)
	assert.Equal(t, types.TypeBoolean, result.Type)
	assert.Equal(t, true, result.Data.(bool))
}
